#include <kernel.h>

/* Would be nice to have one debug output with a mux to console for early init (i.e. before formatters enabled) */
//#define PAGE_DEBUG
//#define PAGE_UPDATE_DEBUG
//#define PTE_DEBUG

#if defined(PAGE_DEBUG) && !defined(BOOT)
#define page_debug(x, ...) do {log_printf("PAGE", "%s: " x, __func__, ##__VA_ARGS__);} while(0)
#else
#define page_debug(x, ...)
#endif

#ifdef KERNEL
static struct spinlock pt_lock;
#define pagetable_lock() u64 _savedflags = spin_lock_irq(&pt_lock)
#define pagetable_unlock() spin_unlock_irq(&pt_lock, _savedflags)
#else
#define pagetable_lock()
#define pagetable_unlock()
#endif

#define PT1 39
#define PT2 30
#define PT3 21
#define PT4 12

static inline u64 flags_from_pte(u64 pte)
{
    return pte & _PAGE_FLAGS_MASK;
}

static heap pageheap;

static const int level_shift[5] = { -1, PT1, PT2, PT3, PT4 };

static u64 pagebase;

static u64 *(*pointer_from_pteaddr)(u64 pa);
static u64 (*pteaddr_from_pointer)(u64 *p);

#ifdef BOOT
static inline u64 *boot_pointer_from_pteaddr(u64 pa)
{
    return (u64*)(u32)pa;
}

static inline u64 boot_pteaddr_from_pointer(u64 *p)
{
    return (u64)(u32)p;
}
#else
static table pt_p2v;
static range pt_initial_phys;

static inline u64 *boot_pointer_from_pteaddr(u64 pa)
{
    return (u64 *)pa;
}

static inline u64 boot_pteaddr_from_pointer(u64 *p)
{
    return (u64)p;
}

static inline u64 *kern_pointer_from_pteaddr(u64 pa)
{
    if (point_in_range(pt_initial_phys, pa)) {
        u64 v = (pa - pt_initial_phys.start) + PAGES_BASE;
        return pointer_from_u64(v);
    }
    u64 offset = pa & MASK(PAGELOG_2M);
    u64 p = pa & ~MASK(PAGELOG_2M);
    u64 v = (u64)table_find(pt_p2v, (void *)p);
    assert(v);
    return pointer_from_u64(v + offset);
}

static physical physical_from_virtual_locked(void *x);
static inline u64 kern_pteaddr_from_pointer(u64 *p)
{
    return physical_from_virtual_locked(p);
}
#endif

static inline u64 pte_lookup_phys(u64 table, u64 vaddr, int offset)
{
    return table + (((vaddr >> offset) & MASK(9)) << 3);
}

#ifndef physical_from_virtual
static inline u64 *pte_lookup_ptr(u64 table, u64 vaddr, int offset)
{
    return pointer_from_pteaddr(pte_lookup_phys(table, vaddr, offset));
}

static inline u64 page_lookup(u64 table, u64 vaddr, int offset)
{
    u64 a = *pte_lookup_ptr(table, vaddr, offset);
    return (a & 1) ? page_from_pte(a) : 0;
}

static physical physical_from_virtual_locked(void *x)
{
    u64 xt = u64_from_pointer(x);
    u64 l3 = page_lookup(pagebase, xt, PT1);
    if (!l3) return INVALID_PHYSICAL;
    u64 l2 = page_lookup(l3, xt, PT2);
    if (!l2) return INVALID_PHYSICAL;
    u64 l1 = page_lookup(l2, xt, PT3); // 2m pages
    if (!l1) return INVALID_PHYSICAL;
    if (*pte_lookup_ptr(l2, xt, PT3) & _PAGE_2M_SIZE) {
        return (l1 | (xt & MASK(PT3)));
    }
    u64 l0 = page_lookup(l1, xt, PT4);
    return l0 ? l0 | (xt & MASK(PT4)) : INVALID_PHYSICAL;
}

physical physical_from_virtual(void *x)
{
    u64 p;
    pagetable_lock();
    p = physical_from_virtual_locked(x);
    pagetable_unlock();
    return p;
}
#endif

/* assumes page table is consistent when called */
void flush_tlb()
{
    u64 *base;
    mov_from_cr("cr3", base);
    mov_to_cr("cr3", base);
}

#ifdef BOOT
void page_invalidate(flush_entry f, u64 address)
{
    flush_tlb();
}

void page_invalidate_sync(flush_entry f, thunk completion)
{
    apply(completion);
}

void page_invalidate_flush()
{

}

flush_entry get_page_flush_entry()
{
    return 0;
}
#endif

#ifndef BOOT
void dump_ptes(void *x)
{
    pagetable_lock();
    u64 xt = u64_from_pointer(x);

    rprintf("dump_ptes 0x%lx\n", x);
    u64 l1 = *pte_lookup_ptr(pagebase, xt, PT1);
    rprintf("  l1: 0x%lx\n", l1);
    if (l1 & 1) {
        u64 l2 = *pte_lookup_ptr(l1, xt, PT2);
        rprintf("  l2: 0x%lx\n", l2);
        if (l2 & 1) {
            u64 l3 = *pte_lookup_ptr(l2, xt, PT3);
            rprintf("  l3: 0x%lx\n", l3);
            if ((l3 & 1) && (l3 & _PAGE_2M_SIZE) == 0) {
                u64 l4 = *pte_lookup_ptr(l3, xt, PT4);
                rprintf("  l4: 0x%lx\n", l4);
            }
        }
    }
    pagetable_unlock();
}
#endif

static void write_pte(u64 target, physical to, u64 flags, boolean * invalidate)
{
    u64 new = to | flags;
    u64 *pteptr = pointer_from_pteaddr(target);
#ifdef PTE_DEBUG
    rputs(", write_pte: ");
    print_u64(u64_from_pointer(target));
    rputs(" = ");
    print_u64(new);
#endif
    assert((new & _PAGE_NO_FAT) == 0);
    if (*pteptr == new) {
#ifdef PTE_DEBUG
	rputs(", pte same; no op");
#endif
	return;
    }
    /* Unconditionally invalidate in case this cpu's TLB is not up-to-date. The
     * TLB can be outdated in SMP situations where another cpu has marked the
     * PTE as not present but this cpu hasn't processed the shootdown yet. */
    *invalidate = true;
#ifdef PTE_DEBUG
    rputs("   invalidate, old ");
    print_u64(*pteptr);
    rputs(", new ");
    print_u64(new);
    rputs("\n");
#endif
    *pteptr = new;
#ifdef PTE_DEBUG
    rputs("\n");
#endif
}

#ifdef PTE_DEBUG
static void print_level(int level)
{
    int i;
    for (i = 0; i < level - 1; i++)
	rprintf(" ");
    rprintf("%d", level);
    for (i = 0; i < 5 - level; i++)
	rprintf(" ");
}
#endif

/* p == 0 && flags == 0 for unmap */
static boolean force_entry(u64 b, u64 v, physical p, int level,
			   boolean fat, u64 flags, boolean *invalidate)
{
    u64 pte_phys = pte_lookup_phys(b, v, level_shift[level]);
    assert((flags & _PAGE_NO_FAT) == 0);

    if (level == (fat ? 3 : 4)) {
#ifdef PTE_DEBUG
	rputs("! ");
	print_level(level);
	rputs(", phys ");
	print_u64(pte_phys);
#endif
	if (fat)
	    flags |= _PAGE_2M_SIZE;
	write_pte(pte_phys, p, flags, invalidate);
	return true;
    } else {
        u64 pte = *pointer_from_pteaddr(pte_phys);
        if (pte & _PAGE_PRESENT) {
            if (pte_is_2M(level, pte)) {
                rputs("\nforce_entry fail: attempting to map a 4K page over an "
                        "existing 2M mapping\n");
                return false;
            }
            /* XXX when unmapping, add a check here to see if the
               directory page is completely unused, and explicitly
               remove and free them when possible. This will avoid the
               occasional invalidate caused by lingering mid
               directories without entries */
	    return force_entry(page_from_pte(pte), v, p, level + 1,
                               fat, flags, invalidate);
	} else {
	    if (flags == 0)	/* only lookup for unmap */
		return false;
	    u64 *n = allocate_zero(pageheap, PAGESIZE);
	    if (n == INVALID_ADDRESS)
		return false;

            /* TODO kind of ridiculous to lose the phys addr from the
               allocation just to look it up again... */
            u64 n_addr = pteaddr_from_pointer(n);
	    if (!force_entry(n_addr, v, p, level + 1, fat, flags, invalidate))
                return false;
#ifdef PTE_DEBUG
            // XXX update debugs
	    rputs("- ");
	    print_level(level);
	    rputs(", phys ");
	    print_u64(pte_phys);
#endif
            /* user and writable are AND of flags from all levels */
	    write_pte(pte_phys, n_addr, _PAGE_WRITABLE | _PAGE_USER | _PAGE_PRESENT, invalidate);
	    return true;
	}
    }
}

/* called with lock held */
static inline boolean map_page(u64 base, u64 v, physical p,
                               boolean fat, u64 flags, boolean * invalidate, flush_entry fe)
{
    boolean invalidate_entry = false;
#ifdef PAGE_UPDATE_DEBUG
    rprintf("force entry base 0x%p, v 0x%lx, p 0x%lx, fat %d, flags 0x%lx\n",
            base, v, p, fat, flags);
#endif
    v &= MASK(VIRTUAL_ADDRESS_BITS);
    if (!force_entry(base, v, p, 1, fat, flags, &invalidate_entry))
	return false;
    if (invalidate_entry) {
        // move this up to construct ranges?
        page_invalidate(fe, v);
        if (invalidate)
            *invalidate = true;
    }
    return true;
}

//#define TRAVERSE_PTES_DEBUG

#define PTE_ENTRIES U64_FROM_BIT(9)
static boolean recurse_ptes(u64 pbase, int level, u64 vstart, u64 len, u64 laddr, entry_handler ph)
{
    int shift = level_shift[level];
    u64 lsize = U64_FROM_BIT(shift);
    u64 start_idx = vstart > laddr ? ((vstart - laddr) >> shift) : 0;
    u64 x = vstart + len - laddr;
    u64 end_idx = MIN(pad(x, lsize) >> shift, PTE_ENTRIES);
    u64 offset = start_idx << shift;

#ifdef TRAVERSE_PTES_DEBUG
    rprintf("   pbase 0x%lx, level %d, shift %d, lsize 0x%lx, laddr 0x%lx,\n"
            "      start_idx %ld, end_idx %ld, offset 0x%lx\n",
            pbase, level, shift, lsize, laddr, start_idx, end_idx, offset);
#endif

    assert(start_idx <= PTE_ENTRIES);
    assert(end_idx <= PTE_ENTRIES);

    for (u64 i = start_idx; i < end_idx; i++, offset += lsize) {
        u64 addr = laddr + (i << shift);
        if (addr & U64_FROM_BIT(47))
            addr |= 0xffff000000000000; /* make canonical */
        u64 pteaddr = pbase + (i * sizeof(u64));
        u64 *pte = pointer_from_pteaddr(pteaddr);
#ifdef TRAVERSE_PTES_DEBUG
        rprintf("   idx %d, offset 0x%lx, addr 0x%lx, pteaddr 0x%lx, *pte %p\n",
                i, offset, addr, pteaddr, *pte);
#endif
        if (!apply(ph, level, addr, pte))
            return false;
        if (!pte_is_present(*pte))
            continue;
        if (level == 3 && (*pte & _PAGE_2M_SIZE) != 0)
            continue;
        if (level < 4) {
            if (!recurse_ptes(page_from_pte(*pte), level + 1, vstart, len,
                              laddr + offset, ph))
                return false;
        }
    }
    return true;
}

boolean traverse_ptes(u64 vaddr, u64 length, entry_handler ph)
{
#ifdef TRAVERSE_PTES_DEBUG
    rprintf("traverse_ptes vaddr 0x%lx, length 0x%lx\n", vaddr, length);
#endif
    pagetable_lock();
    boolean result = recurse_ptes(pagebase, 1, vaddr & MASK(VIRTUAL_ADDRESS_BITS),
                                  length, 0, ph);
    pagetable_unlock();
    return result;
}

/* called with lock held */
closure_function(0, 3, boolean, validate_entry,
                 int, level, u64, vaddr, pteptr, entry)
{
    return pte_is_present(pte_from_pteptr(entry));
}

/* validate that all pages in vaddr range [base, base + length) are present */
boolean validate_virtual(void * base, u64 length)
{
    page_debug("base %p, length 0x%lx\n", base, length);
    return traverse_ptes(u64_from_pointer(base), length, stack_closure(validate_entry));
}

/* called with lock held */
closure_function(2, 3, boolean, update_pte_flags,
                 pageflags, flags, flush_entry, fe,
                 int, level, u64, addr, pteptr, entry)
{
    /* we only care about present ptes */
    pte orig_pte = pte_from_pteptr(entry);
    if (!pte_is_present(orig_pte) || !pte_is_mapping(level, orig_pte))
        return true;

    pte_set(entry, (orig_pte & ~_PAGE_PROT_FLAGS) | bound(flags).w);
#ifdef PAGE_UPDATE_DEBUG
    page_debug("update 0x%lx: pte @ 0x%lx, 0x%lx -> 0x%lx\n", addr, entry, old,
               pte_from_pteptr(entry).w);
#endif
    page_invalidate(bound(fe), addr);
    return true;
}

/* Update access protection flags for any pages mapped within a given area */
void update_map_flags(u64 vaddr, u64 length, pageflags flags)
{
    flags.w &= ~_PAGE_NO_FAT;
    page_debug("vaddr 0x%lx, length 0x%lx, flags 0x%lx\n", vaddr, length, flags.w);
    flush_entry fe = get_page_flush_entry();
    traverse_ptes(vaddr, length, stack_closure(update_pte_flags, flags, fe));
    page_invalidate_sync(fe, ignore);
}

/* called with lock held */
closure_function(3, 3, boolean, remap_entry,
                 u64, new, u64, old, flush_entry, fe,
                 int, level, u64, curr, pteptr, entry)
{
    u64 offset = curr - bound(old);
    u64 oldentry = pte_from_pteptr(entry);
    u64 new_curr = bound(new) + offset;
    u64 phys = page_from_pte(oldentry);
    u64 flags = flags_from_pte(oldentry);
#ifdef PAGE_UPDATE_DEBUG
    page_debug("level %d, old curr 0x%lx, phys 0x%lx, new curr 0x%lx, entry 0x%lx, *entry 0x%lx, flags 0x%lx\n",
               level, curr, phys, new_curr, entry, *entry, flags.w);
#endif

    /* only look at ptes at this point */
    if (!pte_is_present(oldentry) || !pte_is_mapping(level, oldentry))
        return true;

    /* transpose mapped page */
    map_page(pagebase, new_curr, phys, pte_is_2M(level, oldentry), flags, 0, bound(fe));

    /* reset old entry */
    *entry = 0;

    /* invalidate old mapping (map_page takes care of new)  */
    page_invalidate(bound(fe), curr);

    return true;
}

/* We're just going to do forward traversal, for we don't yet need to
   support overlapping moves. Should the latter become necessary
   (e.g. to support MREMAP_FIXED in mremap(2) without depending on
   MREMAP_MAYMOVE), write a "traverse_ptes_reverse" to walk pages
   from high address to low (like memcpy).
*/
void remap_pages(u64 vaddr_new, u64 vaddr_old, u64 length)
{
    page_debug("vaddr_new 0x%lx, vaddr_old 0x%lx, length 0x%lx\n", vaddr_new, vaddr_old, length);
    if (vaddr_new == vaddr_old)
        return;
    assert(range_empty(range_intersection(irange(vaddr_new, vaddr_new + length),
                                          irange(vaddr_old, vaddr_old + length))));
    flush_entry fe = get_page_flush_entry();
    traverse_ptes(vaddr_old, length, stack_closure(remap_entry, vaddr_new, vaddr_old, fe));
    page_invalidate_sync(fe, ignore);
}

/* called with lock held */
closure_function(0, 3, boolean, zero_page,
                 int, level, u64, addr, pteptr, entry)
{
    u64 e = pte_from_pteptr(entry);
    if (pte_is_present(e) && pte_is_mapping(level, e)) {
        u64 size = pte_is_2M(level, e) ? PAGESIZE_2M : PAGESIZE;
#ifdef PAGE_UPDATE_DEBUG
        page_debug("addr 0x%lx, size 0x%lx\n", addr, size);
#endif
        zero(pointer_from_u64(addr), size);
    }
    return true;
}

void zero_mapped_pages(u64 vaddr, u64 length)
{
    traverse_ptes(vaddr, length, stack_closure(zero_page));
}

/* called with lock held */
closure_function(2, 3, boolean, unmap_page,
                 range_handler, rh, flush_entry, fe,
                 int, level, u64, vaddr, pteptr, entry)
{
    range_handler rh = bound(rh);
    u64 old_entry = pte_from_pteptr(entry);
    if (pte_is_present(old_entry) && pte_is_mapping(level, old_entry)) {
#ifdef PAGE_UPDATE_DEBUG
        page_debug("rh %p, level %d, vaddr 0x%lx, entry %p, *entry 0x%lx\n",
                   rh, level, vaddr, entry, *entry);
#endif
        *entry = 0;
        page_invalidate(bound(fe), vaddr);
        if (rh) {
            apply(rh, irangel(page_from_pte(old_entry),
                              (pte_is_2M(level, old_entry) ?
                               PAGESIZE_2M : PAGESIZE)));
        }
    }
    return true;
}

/* Be warned: the page table lock is held when rh is called; don't try
   to modify the page table while traversing it */
void unmap_pages_with_handler(u64 virtual, u64 length, range_handler rh)
{
    assert(!((virtual & PAGEMASK) || (length & PAGEMASK)));
    flush_entry fe = get_page_flush_entry();
    traverse_ptes(virtual, length, stack_closure(unmap_page, rh, fe));
    page_invalidate_sync(fe, ignore);
}

// error processing
static void map_range(u64 virtual, physical p, u64 length, u64 flags)
{
    u64 len = pad(length, PAGESIZE);
    u64 vo = virtual;
    u64 po = p;

    pagetable_lock();
    u64 pb = pagebase;

    /* may be extreme, but can't be careful enough */
    memory_barrier();

    if ((virtual & PAGEMASK) || (p & PAGEMASK) || (length & PAGEMASK)) {
	if (flags == 0)
	    rputs("un");
	rputs("map() called with unaligned paramters!\n v: ");
	print_u64(virtual);
	rputs(", p: ");
	print_u64(p);
	rputs(", length: ");
	print_u64(length);
	halt("\n");
    }

#ifdef PAGE_DEBUG
    rputs("map_range v: ");
    print_u64(virtual);
    rputs(", p: ");
    print_u64(p);
    rputs(", length: ");
    print_u64(length);
    rputs(", flags: ");
    print_u64(flags);
    rputs("\n");
#endif

    boolean invalidate = false;
    flush_entry fe = get_page_flush_entry();
    for (int i = 0; i < len;) {
        boolean fat = ((flags & _PAGE_NO_FAT) == 0) && !(vo & MASK(PT3)) &&
            !(po & MASK(PT3)) && ((len - i) >= (1ull<<PT3));
        if (!map_page(pb, vo, po, fat, flags & ~_PAGE_NO_FAT, &invalidate, fe)) {
            /* may fail if flags == 0 and no mapping, but that's not a problem */
            if (flags)
                halt("map: ran out of page table memory\n");
        }
        int off = 1ull << (fat ? PT3 : PT4);
        vo += off;
        po += off;
        i += off;
    }
#ifdef PAGE_DEBUG
    if (invalidate && p)        /* don't care about invalidate on unmap */
        rputs("   - part of map caused invalidate\n");
#endif

    memory_barrier();
    pagetable_unlock();
    page_invalidate_sync(fe, ignore);
}

void map(u64 virtual, physical p, u64 length, pageflags flags)
{
    map_range(virtual, p, length, flags.w | _PAGE_PRESENT);
}

void unmap(u64 virtual, u64 length)
{
#ifdef PAGE_DEBUG
    rputs("unmap v: ");
    print_u64(virtual);
    rputs(", length: ");
    print_u64(length);
    rputs("\n");
#endif
    unmap_pages(virtual, length);
}

void *bootstrap_page_tables(heap initial)
{
    /* page table setup */
    pageheap = initial;
    void *pgdir = allocate_zero(initial, PAGESIZE);
    assert(pgdir != INVALID_ADDRESS);
    pagebase = u64_from_pointer(pgdir);
    pointer_from_pteaddr = boot_pointer_from_pteaddr;
    pteaddr_from_pointer = boot_pteaddr_from_pointer;
#ifdef KERNEL
    spin_lock_init(&pt_lock);
#endif
    return pgdir;
}

#ifdef KERNEL
static id_heap phys_internal;
static u64 pt_2m_next;

closure_function(0, 1, void, dealloc_phys_page,
                 range, r)
{
    if (!id_heap_set_area(phys_internal, r.start, range_span(r), true, false))
        msg_err("some of physical range %R not allocated in heap\n", r);
}

void unmap_and_free_phys(u64 virtual, u64 length)
{
    unmap_pages_with_handler(virtual, length, stack_closure(dealloc_phys_page));
}

/* pt_lock should already be held here */
static u64 pt_2m_alloc(heap h, bytes size)
{
    /* carve out a virtual ... no dealloc here - we don't dealloc pt
       pages either, but if we wanted to, the child 4k page heap
       will take care of reuse */
    assert(pt_2m_next > 0);
    u64 len = pad(size, PAGESIZE_2M);
    u64 v = pt_2m_next;
    pt_2m_next += len;
    assert(pt_2m_next >= PAGES_BASE);

    flush_entry fe = get_page_flush_entry();
    for (u64 i = v; i < v + len; i += PAGESIZE_2M) {
        u64 p = allocate_u64((heap)phys_internal, PAGESIZE_2M);
        if (p == INVALID_PHYSICAL)
            halt("%s: failed to allocate 2M physical page\n", __func__);
        /* we depend the pmd already being installed to avoid an alloc here */
        map_page(pagebase, i, p, true, _PAGE_WRITABLE | _PAGE_PRESENT, 0, fe);
        table_set(pt_p2v, (void *)p, (void *)i);
    }
    page_invalidate_sync(fe, ignore);
    return v;
}

void map_setup_2mbpages(u64 v, physical p, int pages, pageflags flags,
                        u64 *pdpt, u64 *pdt)
{
    assert(!(v & PAGEMASK_2M));
    assert(!(p & PAGEMASK_2M));
    u64 *pml4;
    mov_from_cr("cr3", pml4);
    flags.w |= _PAGE_PRESENT;
    pml4[(v >> PT1) & MASK(9)] = u64_from_pointer(pdpt) | flags.w;
    v &= MASK(PT1);
    pdpt[v >> PT2] = u64_from_pointer(pdt) | flags.w;
    v &= MASK(PT2);
    assert(v + pages <= 512);
    for (int i = 0; i < pages; i++)
        pdt[v + i] = (p + (i << PAGELOG_2M)) | flags.w | _PAGE_2M_SIZE;
    memory_barrier();
}

/* this happens even before moving to the new stack, so ... be cool */
void init_page_tables(heap h, id_heap physical, range initial_phys)
{
    write_msr(EFER_MSR, read_msr(EFER_MSR) | EFER_NXE);
    mov_from_cr("cr3", pagebase);
    pointer_from_pteaddr = kern_pointer_from_pteaddr;
    pteaddr_from_pointer = kern_pteaddr_from_pointer;
    spin_lock_init(&pt_lock);
    phys_internal = physical;

    pt_p2v = allocate_table(h, identity_key, pointer_equal);
    assert(pt_p2v != INVALID_ADDRESS);

    /* store initial boundaries for p->v lookup */
    pt_initial_phys = initial_phys;
    pt_2m_next = PAGES_BASE + pad(range_span(initial_phys), PAGESIZE_2M);

    /* 2m heap for stage3 pt allocs */
    heap pt_2m = allocate(h, sizeof(struct heap));
    assert(pt_2m != INVALID_ADDRESS);
    pt_2m->alloc = pt_2m_alloc;
    pt_2m->dealloc = leak;
    pt_2m->pagesize = PAGESIZE_2M;

    /* 4k page heap */
    pageheap = (heap)create_id_heap_backed(h, /* XXX */ h, pt_2m, PAGESIZE, true);
    assert(pageheap != INVALID_ADDRESS);
}
#else
/* stage2 */
void init_page_tables(heap initial)
{
    void *vmbase = bootstrap_page_tables(initial);
    mov_to_cr("cr3", vmbase);
}
#endif
